﻿using System;
using System.Linq;
using System.Data.Common;
using System.Data.SqlClient;
using System.Collections.Generic;

using Platform.Extensions;

using Microsoft.EntityFrameworkCore;
using Microsoft.Extensions.Logging;
using System.Data;

namespace Platform.Sql.ChangeMonitor
{
    public class QueryChangeMonitor : IQueryChangeMonitor
    {
        #region Construction and fields

        // static list of all monitors
        private static List<QueryChangeMonitor> _Monitors
            = new List<QueryChangeMonitor>();

        // current monitor proerties
        protected readonly ILogger Logger;
        protected SqlDependency Dependency;
        protected string ConnectionString;
        protected string Sql;

        public QueryChangeMonitor(ILogger<QueryChangeMonitor> logger)
        {
            Logger = logger;
        }

        ~QueryChangeMonitor()
        {
            Dispose(false);
        }

        public void Dispose()
        {
            Dispose(true);
            GC.SuppressFinalize(this);
        }

        protected virtual void Dispose(bool disposing)
        {
            if (disposing)
            {
                Changed = null;
                Stop();
            }
        }

        #endregion

        #region Properties

        /// <summary>
        /// Gets or sets the query change monitor name
        /// </summary>
        public string Name
        {
            get;
            set;
        }

        #endregion

        #region Start & Stop

        /// <summary>
        /// Start monitor query
        /// </summary>
        /// <typeparam name="TEntity"></typeparam>
        /// <param name="query"></param>
        public void Start<TEntity>(DbContext context, IQueryable<TEntity> query)
            where TEntity : class
        {
            Start(context, query.ToSql());
        }

        /// <summary>
        /// Start monitor query
        /// </summary>
        /// <typeparam name="TEntity"></typeparam>
        /// <param name="query"></param>
        public void Start(DbContext context, string sql)
        {
            Start(context.Database.GetDbConnection().ConnectionString, sql);
        }

        /// <summary>
        /// Start monitor query
        /// </summary>
        /// <param name="connection"></param>
        /// <param name="sql"></param>
        public void Start(string connection, string sql)
        {
            lock (_Monitors)
            {
                // add ourselfs
                if (_Monitors.Contains(this))
                    return;

                // set properties
                ConnectionString = connection;
                Sql = sql;

                // start
                if (!_Monitors.Any(m => m.ConnectionString == ConnectionString))
                    SqlDependency.Start(ConnectionString);

                // add to monitors
                _Monitors.Add(this);

                // start notifications
                StartNotification();
            }
        }

        /// <summary>
        /// Stop monitor query
        /// </summary>
        public void Stop()
        {
            lock (_Monitors)
            {
                // remove ourselfs
                if (!_Monitors.Contains(this))
                    return;

                // stop notifications
                StopNotification();

                // remove from monitors
                _Monitors.Remove(this);

                // stop
                if (!_Monitors.Any(m => m.ConnectionString == ConnectionString))
                    SqlDependency.Stop(ConnectionString);
            }
        }

        #endregion

        #region Change Notification

        /// <summary>
        /// Register for change notifications
        /// </summary>
        public event EventHandler Changed;

        /// <summary>
        /// Raise change event
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        protected virtual void OnChanged(object sender, EventArgs e)
        {
            Changed?.Invoke(sender, e);
        }

        /// <summary>
        /// Start notifications
        /// </summary>
        private void StartNotification()
        {
            using (var connection = new SqlConnection(ConnectionString))
            {
                connection.Open();

                using (var command = new SqlCommand(Sql, connection))
                {

                    Dependency = new SqlDependency(command);
                    Dependency.OnChange += OnDependencyChanged;
                    int count = command.ExecuteNonQuery();
                }

                // connection.Close();
            }

            OnNotificationReload();
        }

        /// <summary>
        /// Stop notifications
        /// </summary>
        private void StopNotification()
        {
            Dependency.OnChange -= OnDependencyChanged;
            Dependency = null;
        }

        /// <summary>
        /// Process Dependency Changes
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void OnDependencyChanged(object sender, SqlNotificationEventArgs e)
        {
            Logger?.LogDebug(new EventId(1, "SqlDependencyChanged"), "Sql Dependency: {0}, {1}, {2}: {3}", e.Info, e.Source, e.Type, Sql);

            switch (e.Info)
            {
                case SqlNotificationInfo.Invalid:
                    Logger?.LogError(new EventId(2, "SqlDependencyInvalid"), "Sql Dependency failed: {0}, {1}, {2}: {3}", e.Info, e.Source, e.Type, Sql);
                    return;
            }

            if (e.Type == SqlNotificationType.Subscribe)
                lock (this)
                {
                    StopNotification();
                    StartNotification();
                    return;
                }

            if (e.Type != SqlNotificationType.Change)
            {
                Logger?.LogInformation(new EventId(3, "SqlDependencyNotChange"), "Sql Dependency no change: {0}, {1}, {2}: {3}", e.Info, e.Source, e.Type, Sql);
                return;
            }

            OnChanged(this, EventArgs.Empty);

            lock (_Monitors)
            {
                StopNotification();
                StartNotification();
            }
        }

        protected virtual void OnNotificationReload()
        {
        }

        #endregion
    }
}
